# coding: utf-8

"""This module contains the code specific to applications at the LaMcube
laboratory.

They are included in Crappy as most users are from the LaMcube, but kept in a
separate submodule because they're almost certainly never going to be used
as-is by anyone outside the lab.
"""

from .biaxe import Biaxe
from .bispectral import BiSpectral
from .biotens import Biotens
